program lvr;

///////////////////////////////////////////////////////////////////////////////
//
// Title : lazarus versioninfo renamer (lvr) v0.1 alpha
// Author: TRon
// Date  : 8 feb 2023
//
///////////////////////////////////////////////////////////////////////////////
//
// Description:
// Inspired by the question from MountainQ on how to use version information
// from a Lazarus project settings for usage in the filename of the created
// executable.
//
// Slightly based on code from user Ally
//
// references:
// - original thread: https://forum.lazarus.freepascal.org/index.php/topic,62206.0.html
// - Lazarus macros : https://wiki.lazarus.freepascal.org/IDE_Macros_in_paths_and_filenames
//
// Disclaimer:
// Use this example code at your own risk. The example code is to be used only
// to show /how/ a tool could be written/used to extract build version
// information from an executable and use that extracted information to
// copy/rename the original executable file.
//
// This example contains no proper error-checking and no attention whatsoever
// was given to safety and/or correct workings.
//
// This example can/will (re)move/copy/delete/destruct files, may harm your
// computer, may contribute to global warming and contains bad coding style.
//
// This examle code is intended to be used by Lazarus only.
//
// Notes:
// - Files that have a file extension are currently not supported (correctly).
//
// License:
// Pay-It-Forward
//
///////////////////////////////////////////////////////////////////////////////
//
// Installation:
// Compile this source and place the generated executable to a dedicated
// location.
//
// lazarus configuration:
// - Create a new project
// - View your project option
// - Select project options, category compiler commands
// - at the right panel subpanel Execute After check the desired calls for which
//   you wish to invoke lvr, e.g. compile/build/run
// - In the command edit box add lvr and its arguments, e.g.:
//   /path/to/lvr $(OutputFile) "%(filename)-%(major).%(minor).%(revision).%(build)"
// - inside the parsers checklistbox, check the option "show all outputlines"
// - Select category project options, Version Info
// - check Include Version info in executable
// - change the Major version, Minor version, Revision and Build so that they
//   contain some sane numbers.
// - Press ok on the projects options dialog.
// - Save your project
// - compile/build/run your project (based on the checks at "execute after" that
//   were made) and lvr should be invoked.
// - Look at the message window to see the output from the lvr command
//   Typical output looks like:
//   Compile Project, Target: project1: Success
//   Project: Executing command after: Success
//   lazarus versioninfo renamer v0.1 alpha
//   File /path/to/your/project copied to /path/to/your/project-1.2.3.4
//
// quick usage:
// Project Options/Compiler Options/Compiler Commands/Execute After/Command:
// /path/to/lvr $(OutputFile) "%(filename)-%(major).%(minor).%(revision).%(build)"
// macro names are optional and can be placed anywhere in the second argument
//
///////////////////////////////////////////////////////////////////////////////

{$mode objfpc}{$H+}
{$modeswitch typehelpers}{$modeswitch arrayoperators}

// activate following define for actual renaming (=removal of original file)
// Note that when you do so you come into problems when you want to use the
// run/debug command as the original executable is not available anymore (so
// lazarus will be confused, e.g.
// No program file "/path/to/yor/projects/executable" found.).
{.$define copy_and_delete_file}

// Activate following define to have lvr output some additonal information
// that might be helpful in case of troubles.
{.$define blab}

uses
  baseunix, classes, sysutils, fileinfo,
  winpeimagereader, elfreader, machoreader;

type
  TFilenameHelper = type helper for TFilename
    function  GetFileName: string;
    function  GetFilePath: string;
    function  GetFileDir: string;
    property  FileDir: string read GetFileDir;
    property  FilePath: string read GetFilePath;
    property  FileName: string read GetFileName;
  end;

function TFilenameHelper.GetFileDir: string;
begin
  result := ExtractFileDir(Self);
end;

function TFilenameHelper.GetFilePath: string;
begin
  result := ExtractFilePath(Self);
end;

function  TFilenameHelper.GetFileName: string;
begin
  result:=ExtractFilename(Self); //  result:=ExtractFilenameOnly(Self);
end;

function CopyFile(from_name: string; to_name: string): boolean;
var
  SrcFS, DstFS: TFileStream;
begin
  // layman's copy function: /never/ use in production code
  SrcFS:= TFileStream.Create(from_name, fmOpenRead);
  DstFS:= TFileStream.Create(to_name, fmCreate);
  DstFS.CopyFrom(SrcFS, SrcFS.Size);
  SrcFS.Free;
  DstFS.Free;
  // wrongfully assume that commands issued above executed well
  result:=true;
end;

function RenameFile(from_name: string; to_name: string): boolean;
begin
  // copy and delete file (because RenameFile() does not work consistent amongst platforms)
  CopyFile(from_name, to_name);
  DeleteFile(from_name);
  // wrongfully assume that commands issued above executed well
  result:= true;
end;

function GetProgramVersion(Filename: string; out ProgramVersion: TProgramVersion): boolean;
// shameless copy from unit fileinfo with small addition
var
  VersionInfo: TVersionInfo = nil;
begin
  result:= false;
  if FileExists(Filename) then
  begin
    VersionInfo:= TVersionInfo.Create;
    try
      try
        VersionInfo.Load(Filename);
        ProgramVersion.Major    := VersionInfo.FixedInfo.FileVersion[0];
        ProgramVersion.Minor    := VersionInfo.FixedInfo.FileVersion[1];
        ProgramVersion.Revision := VersionInfo.FixedInfo.FileVersion[2];
        ProgramVersion.Build    := VersionInfo.FixedInfo.FileVersion[3];
        result:= true;
      except
        // ignore
      end;
    finally
      VersionInfo.Free;
    end;
  end;  // fileexists
end;

function VersionFormat(Fmt: String; ProgramVersion: TProgramVersion; from_Filename: TFilename): string;
const
  rf = [rfReplaceAll, rfIgnoreCase];
var
  ret : string absolute result;
begin
  // replace all format parameters to strings and return
  ret := fmt;
  ret := ret.replace('%(major)'   , ProgramVersion.Major.ToString    , rf);
  ret := ret.replace('%(minor)'   , ProgramVersion.Minor.ToString    , rf);
  ret := ret.replace('%(revision)', ProgramVersion.Revision.ToString , rf);
  ret := ret.replace('%(build)'   , ProgramVersion.Build.ToString    , rf);

  ret := ret.replace('%(filepath)', from_filename.FilePath, rf);
  ret := ret.replace('%(filename)', from_filename.FileName, rf);
end;

function HandleOptions(out SFilename, DFilename: string): integer;
// the most simplistic and dumb argument handler ever
// first argument /must/ point to the original executable (including path)
// second argument /must/ to the destination name (path may be omitted and
// name may contain special identifiers that are replaced automatically)
begin
  SFileName := '';
  DFileName := '';
  {$ifdef blab}
  writeln('paramcount = ', Paramcount);
  {$endif}
  if ParamCount = 2 then
  begin
    result := ParamCount;
    SFileName := ParamStr(1);
    DFileName := ParamStr(2);
    {$ifdef blab}
    writeln('ParamStr(1) = ', SFilename);
    writeln('ParamStr(2) = ', DFilename);
    {$endif}
  end
  else Result := -1;
end;

var
  StatusString: string  = '';
  StatusCode  : integer = -1;

procedure SetStatus(errorcode: integer; message: string); override;
begin
  StatusString := message;
  StatusCode   := errorcode;
end;

procedure SetStatus(errorcode: integer; message: string; const params: array of const); override;
begin
  StatusString := Format(message, params);
  StatusCode   := errorcode;
end;

procedure Main;
var
  OriginalFilename : TFilename = '';
  RenamedFormat    : string = '';
  RenamedFilename  : TFilename = '';
  ProgramVersion   : TProgramVersion;
begin
  if HandleOptions(OriginalFileName, RenamedFormat) = 2 then
  begin
    // refuse to continue when original filename does not contain a path
    if OriginalFileName.FileDir = '' then
    begin
      SetStatus(-20, 'First argument %s does not contain a filename with a full path', [OriginalFilename]);
      exit;
    end;

    if GetProgramVersion(OriginalFilename, ProgramVersion) then
    begin
      RenamedFilename := VersionFormat(RenamedFormat, ProgramVersion, OriginalFileName);

      // add full path from original filename in case not defined in destination name commandline option
      if RenamedFilename.FileDir = ''
        then RenamedFilename :=  OriginalFilename.FilePath + RenamedFileName;

      {$ifdef blab}
      writeln('original filename: "', OriginalFileName,'"');
      writeln('renamed  filename: "', RenamedFileName ,'"');
      {$endif}

      {$ifdef copy_and_delete_file}
      if RenameFile(OriginalFilename, RenamedFilename)
        then SetStatus(100, 'File %s renamed to %s', [OriginalFilename, RenamedFilename])
          else SetStatus(-100, 'Renaming of file %s failed', [OriginalFilename]);
      {$else}
      if CopyFile(OriginalFilename, RenamedFilename)
        then SetStatus(200, 'File %s copied to %s', [OriginalFilename, RenamedFilename])
          else SetStatus(-200, 'Copying of file %s failed', [OriginalFilename]);
      {$endif}
    end
    else SetStatus(-30, 'Unable to obtain version information from file %s', [OriginalFilename]);
  end
  else SetStatus(-10, 'Wrong number of arguments');
end;

begin
  WriteLn('lazarus versioninfo renamer v0.1 alpha');
  try
    Main;
  finally
    case StatusCode of
     100, 200 : exitCode := 0;
     else       exitCode := StatusCode * -1;
    end;
    writeln(StatusString);
  end;
end.